#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

#define MAX_BUFFER 10
#define NAME "easy"
#define AUTHOR "Andrea Marchesini - bakunin@autistici.org"

typedef struct buffer
{
  char value[MAX_BUFFER];
  int id;

  char *name;
  char (*pop) (struct buffer *);
  char (*push) (struct buffer *, char);
  char (*add) (struct buffer *, char);
  char (*get) (struct buffer *);

  struct buffer *next;
} buffer;

buffer *buffers = NULL;
int line=0;

int get_item(FILE *fl, char *where) {
  char buf[1024];
  int i=0;
  char ch;

  while(1) {
    if((ch=getc(fl))==EOF) return 0;

    if(ch=='#') {
      while((ch=getc(fl))!=EOF && ch!='\n');
      line++;
    }

    if(ch=='\n') line++;

    if(ch=='\\') {
      if((ch=getc(fl))==EOF) return 0;
      buf[i++]=ch;
    }

    else if(ch=='\'') {
      buf[i++]=ch;
      if((ch=getc(fl))==EOF) return 0;
      if(ch=='\n') line++;
      buf[i++]=ch;
      if((ch=getc(fl))==EOF || ch!='\'') return 0;
      buf[i++]=ch;
    }

    else if(ch!=' ' && ch!='\t' && ch!='\v' && ch!='\n') {
      buf[i++]=ch;
    }

    else if(i) {
      buf[i++]=0;
      memcpy(where, buf, i);
      return 1;
    }
  }

  return 0;
}

void
panic (char *str, ...)
{
  va_list va;

  va_start (va, str);
  fprintf (stderr, "\nLine %d: ", line-1);
  vfprintf (stderr, str, va);
  fprintf (stderr, "\n");
  va_end (va);

  exit (1);
}

int
is_number (char *a)
{
  int i;
  int len = strlen (a);

  if ((a[0] < '0' || a[0] > '9') && a[0] != '-')
    return 0;

  for (i = 1; i < len; i++)
    if (a[i] < '0' || a[i] > '9')
      return 0;
  return 1;
}

int
is_char (char *a)
{
  if (strlen (a) != 3)
    return 0;
  if (a[0] != '\'')
    return 0;
  if (a[2] != '\'')
    return 0;

  return 1;
}

char
generic_pop (buffer * b)
{
  char v;

  if (!b->id)
    panic ("Buffer %s is empty!", b->name);
  b->id--;
  v = b->value[b->id];

  return v;
}

char
generic_push (buffer * b, char v)
{
  if (b->id >= MAX_BUFFER - 1)
    panic ("Overflow of buffer %s!", b->name);
  b->value[b->id] = v;
  b->id++;

  return 0;
}

char
generic_get (buffer * b)
{
  if (!b->id)
    return 0;
  return b->value[b->id - 1];
}

char
generic_add (buffer * b, char v)
{
  if (!b->id)
    panic ("Buffer %s is empty!", b->name);
  b->value[b->id - 1] += v;

  return 0;
}

char
return_0 (buffer * p)
{
  return 0;
}

void
buffers_new (char *b, void *pop, void *push, void *add, void *get)
{
  buffer *n;

  if (!(n = (buffer *) malloc (sizeof (buffer))))
    panic ("Memory");

  memset (n, 0, sizeof (buffer));
  n->pop = pop ? pop : generic_pop;
  n->push = push ? push : generic_push;
  n->add = add ? add : generic_add;
  n->get = get ? get : generic_get;
  n->next = buffers;
  buffers = n;
  n->name = strdup (b);
}

void
buffers_del (char *b)
{
  buffer *n, *old = NULL;

  n = buffers;

  while (n)
    {
      if (!strcmp (n->name, b))
	{
	  if (!old)
	    {
	      buffers = buffers->next;
	      free (n->name);
	      free (n);
	      break;

	    }
	  else
	    {
	      old->next = n->next;
	      free (n->name);
	      free (n);
	      return;
	    }
	}

      old = n;
      n = n->next;
    }
}

void
buffers_set (char *b1, char *b2)
{
  buffer *n;
  char a;

  if (is_number (b1))
    a = (char) atoi (b1);

  else if (is_char (b1))
    a = b1[1];

  else
    {
      n = buffers;
      while (n)
	{
	  if (!strcmp (n->name, b1))
	    {
	      a = n->pop (n);
	      break;
	    }

	  n = n->next;
	}

      if (!n)
	panic ("No buffer %s founded!", b1);
    }

  n = buffers;
  while (n)
    {
      if (!strcmp (n->name, b2))
	{
	  n->push (n, a);
	  break;
	}

      n = n->next;
    }

  if (!n)
    panic ("No buffer %s founded!", b2);
}

char
buffers_get (char *b)
{
  buffer *n;

  n = buffers;
  while (n)
    {
      if (!strcmp (n->name, b))
	{
	  return n->get (n);
	}

      n = n->next;
    }

  if (!n)
    panic ("No buffer %s founded!", b);

  return 0;
}

void
buffers_ugual (char *b1, char *b2)
{
  buffer *n;
  char a;

  if (is_number (b2))
    a = (char) atoi (b2);

  else if (is_char (b2))
    a = b2[1];

  else
    {
      n = buffers;
      while (n)
	{
	  if (!strcmp (n->name, b2))
	    {
	      a = n->pop (n);
	      n->push(n, a);
	      break;
	    }

	  n = n->next;
	}

      if (!n)
	panic ("No buffer %s founded!", b2);
    }

  n = buffers;
  while (n)
    {
      if (!strcmp (n->name, b1))
	{
	  n->push(n, a);
	  break;
	}

      n = n->next;
    }

  if (!n)
    panic ("No buffer %s founded!", b1);
}

void
buffers_add (char *b1, char *b2)
{
  buffer *n;
  char a;

  if (is_number (b2))
    a = (char) atoi (b2);

  else if (is_char (b2))
    a = b2[1];

  else
    {
      n = buffers;
      while (n)
	{
	  if (!strcmp (n->name, b2))
	    {
	      a = n->pop (n);
	      break;
	    }

	  n = n->next;
	}

      if (!n)
	panic ("No buffer %s founded!", b2);
    }

  n = buffers;
  while (n)
    {
      if (!strcmp (n->name, b1))
	{
	  n->add (n, a);
	  break;
	}

      n = n->next;
    }

  if (!n)
    panic ("No buffer %s founded!", b1);
}

char
write_data (buffer * b, char v)
{
  write (STDIN_FILENO, &v, 1);
  return v;
}

char
null_data (buffer * b, char v)
{
  return v;
}

int
read_data (buffer * b)
{
  char value;

  read (STDIN_FILENO, &value, sizeof (value));

  return value;
}

int
compare_data (buffer * b)
{
  int id;
  int v;

  if (!b->id)
    return 1;

  v = b->value[0];

  for (id = 1; id < b->id; id++)
    {
      if (b->value[id] != v)
	return 1;
    }

  return 0;
}

int
buffers_compare (char *b)
{
  buffer *n;

  n = buffers;
  while (n)
    {
      if (!strcmp (n->name, b))
	return compare_data (n);

      n = n->next;
    }

  if (!n)
    panic ("No buffer %s founded!", b);

  return 0;
}

void
buffers_empty (char *a, char *b)
{
  buffer *n;

  n = buffers;
  while (n)
    {
      if (!strcmp (n->name, a))
	{
	  if (n->id)
	    buffers_set ("1", b);
	  else
	    buffers_set ("0", b);
	  break;
	}

      n = n->next;
    }

  if (!n)
    panic ("No buffer %s founded!", a);
}

int
main (int argc, char **argv)
{
  FILE *fl;

  char cmd[MAX_BUFFER];
  char item1[MAX_BUFFER];
  char item2[MAX_BUFFER];
  int parentesis = 0;
  int quadrates = 0;
  int elsed = 0;
  long cicle_offset = 0;
  char *cicle_value = NULL;

  if (argc == 2) {
    if (!(fl = fopen (argv[1], "r")))
      panic ("Opening %s error.", argv[1]);

  } else if(argc == 1) {
	  fl=stdin;

  } else {
    panic (NAME " - " AUTHOR "\n\nUsage:\t%s file.", argv[0]);
  }

  buffers_new ("i", read_data, null_data, null_data, return_0);
  buffers_new ("o", return_0, write_data, write_data, return_0);
  buffers_new ("n", return_0, null_data, null_data, return_0);

  while (!feof (fl))
    {
      if (!get_item(fl, item1)) {
	panic ("Syntax error!");
	break;
      }

      if (!strcmp (item1, "]"))
	{
	  if (quadrates)
	    {
	      if (cicle_value && buffers_get (cicle_value))
		fseek (fl, cicle_offset, SEEK_SET);

	      else if (cicle_value)
		{
		  free (cicle_value);
		  cicle_value = NULL;
		  quadrates--;
		}

	      continue;
	    }
	  else
	    panic ("Syntax error: why ] ?");
	}

      if (!strcmp (item1, ")"))
	{
	  if (parentesis)
	    {
	      int p = 0;

	      parentesis--;

	      if (elsed)
		{
		  elsed = 0;
		  continue;
		}

	      elsed = 0;

	      if (!get_item(fl, item1))
		panic ("Syntax error: need |");

	      if (strcmp (item1, "|"))
		panic ("Syntax error: need |");

	      if (!get_item(fl, item1))
		panic ("Syntax error: need (");

	      if (strcmp (item1, "("))
		panic ("Syntax error: need (");

	      while (get_item(fl, item1))
		{
		  if (!strcmp (item1, "("))
		    p++;
		  else if (!strcmp (item1, ")"))
		    {
		      if (!p)
			break;
		      else
			p--;
		    }
		}

	      continue;
	    }
	  else
	    panic ("Syntax error: no axpect )");
	}

      if (!strcmp (item1, "#"))
	return 0;

      if (!get_item (fl, cmd))
	panic ("Syntax error: no command ?");

      if (strlen (cmd) > 1)
	panic ("Syntax error: command too long!");

      switch (cmd[0])
	{
	case '>':
	  if (!get_item(fl, item2))
	    panic ("Syntax error: no second operator for >");
	  buffers_set (item1, item2);
	  break;

	case '<':
	  if (!get_item (fl, item2))
	    panic ("Syntax error: no second operator for <");
	  buffers_set (item2, item1);
	  break;

	case '+':
	  if (!get_item(fl, item2))
	    panic ("Syntax error: no second operator for +");
	  buffers_add (item1, item2);
	  break;

	case '=':
	  if (!get_item (fl, item2))
	    panic ("Syntax error: no second operator for =");
	  buffers_ugual (item1, item2);
	  break;

	case '^':
	  buffers_new (item1, NULL, NULL, NULL, NULL);
	  break;

	case '_':
	  buffers_del (item1);
	  break;

	case '?':
	  if (!get_item(fl, item2))
	    panic ("Syntax error: no second operator for ?");
	  buffers_empty (item1, item2);
	  break;

	case ':':
	  if (!get_item (fl, item2))
	    if (strcmp (item2, "["))
	      panic ("Syntax error: need [ after :");

	  if (buffers_get (item1))
	    {
	      quadrates++;
	      cicle_offset = ftell (fl);
	      if (cicle_value)
		panic ("Syntax error: you can't do more one cicles!");
	      cicle_value = strdup (item1);

	    }
	  else
	    {
	      int p = 0;

	      while (get_item (fl, item2))
		{
		  if (!strcmp (item2, "["))
		    p++;
		  else if (!strcmp (item2, "]"))
		    {
		      if (!p)
			break;
		      else
			p--;
		    }
		}
	    }

	  break;

	case '~':
	  if (!get_item (fl, item2))
	    if (strcmp (item2, "("))
	      panic ("Syntax error: need ( after ~");

	  parentesis++;

	  if (!buffers_compare (item1))
	    elsed = 0;

	  else
	    {
	      int p = 0;
	      elsed = 1;

	      while (get_item (fl, item2))
		{
		  if (!strcmp (item2, "("))
		    p++;
		  else if (!strcmp (item2, ")"))
		    {
		      if (!p)
			break;
		      else
			p--;
		    }
		}

	      if (!get_item (fl, item1))
		panic ("Syntax error: need |");

	      if (strcmp (item1, "|"))
		panic ("Syntax error: need |");

	      if (!get_item (fl, item1))
		panic ("Syntax error: need (");

	      if (strcmp (item1, "("))
		panic ("Syntax error: need (");

	    }

	  break;

	default:
	  panic ("Syntax error: %s ?!?", cmd);
	}
    }

  return 0;
}
